using FFTW
using Random # only required to control random seed

function ft_phase_screen(N, r0=0.1, seed=0; Nsh=0, pythonrng=false)

    rng = Xoshiro(seed)
    s = randn(rng, ComplexF64, N, N)
    
    # side length of phase screen
    D = 1
    # outer scale of turbulence
    L0 = 1e2
    
    # angular spatial frequency vectors
    dκ = 2π/D
    κ = fftfreq(N, N) .* dκ

    c = sqrt(2 * 0.49 * r0^(-5/3)) * dκ
    for (m, n) ∈ Iterators.product(axes(s)...)
        # assume von Karman spectrum, factor two for correct variance in real and imaginary parts of the two output screens
        s[m, n] *= c * (κ[m]^2 + κ[n]^2 + (2π/L0)^2)^(-11/12)
    end
    s[1, 1] = 0

    screen = real(bfft!(s))

    if Nsh > 0
        dx = D/N
        # preallocate spatial grid arrays
        x = fftfreq(N, N * dx) |> fftshift
        ssh = zeros(ComplexF64, 3, 3)
        
        # loop over frequency grids with spacing 1/(q^p*L)
        for p ∈ Base.OneTo(Nsh)
            # subharmonic frequency grid [rad/m] 
            dκ = 2π / D / (3^p)
            κsh = fftfreq(3, 3 * dκ)
            
            c = sqrt(2 * 0.49 * r0^(-5/3)) * dκ
            # myrandn!(rng, ssh; pythonrng)
            randn!(rng, ssh)
            
            for (n, m) ∈ Iterators.product(1:3, 1:3)
                # assume von Karman spectrum, factor two for correct variance in real and imaginary parts of the two output screens
                ssh[m, n] *= c * (κsh[m]^2 + κsh[n]^2 + (2π/L0)^2)^(-11/12)
                
                # loop over frequencies on subharmonic grid
                for (i, j) ∈ Iterators.product(axes(screen)...)
                    screen[i, j] += real(ssh[m, n] * cis(κsh[m] * x[i] + κsh[n] * x[j]))
                end
            end
        end
    end

    return screen
end

function main(N=256; Nsh=1)
    nruns = 100
    stats = @timed for i in range(1, nruns)
        ft_phase_screen(N; Nsh)
    end

    display("Runtime (ms): " * string(stats.time / nruns * 1e3)) 

    return
end

main()